@@ -33,6 +33,7 @@ from fortls.constants import (
 )
 from fortls.helper_functions import (
     expand_name,
+    fortran_md,
     get_line_prefix,
     get_paren_level,
     get_var_stack,
@@ -40,24 +41,27 @@ from fortls.helper_functions import (
     resolve_globs,
     set_keyword_ordering,
 )
-from fortls.intrinsics import (
+from fortls.json_templates import change_json, symbol_json, uri_json
+from fortls.jsonrpc import JSONRPC2Connection, path_from_uri, path_to_uri
+from fortls.parsers.internal.ast import FortranAST
+from fortls.parsers.internal.imports import Import
+from fortls.parsers.internal.intrinsics import (
     Intrinsic,
     get_intrinsic_keywords,
     load_intrinsics,
     set_lowercase_intrinsics,
 )
-from fortls.json_templates import change_json, symbol_json, uri_json
-from fortls.jsonrpc import JSONRPC2Connection, path_from_uri, path_to_uri
-from fortls.objects import (
-    FortranAST,
-    Variable,
+from fortls.parsers.internal.parser import FortranFile, get_line_context
+from fortls.parsers.internal.scope import Scope
+from fortls.parsers.internal.use import Use
+from fortls.parsers.internal.utilities import (
     climb_type_tree,
     find_in_scope,
     find_in_workspace,
     get_use_tree,
 )
-from fortls.parse_fortran import FortranFile, get_line_context
-from fortls.regex_patterns import src_file_exts
+from fortls.parsers.internal.variable import Variable
+from fortls.regex_patterns import create_src_file_exts_str
 from fortls.version import __version__
 
 # Global regexes
@@ -85,7 +89,9 @@ class LangServer:
 
         self.sync_type: int = 2 if self.incremental_sync else 1
         self.post_messages = []
-        self.FORTRAN_SRC_EXT_REGEX: Pattern[str] = src_file_exts()
+        self.FORTRAN_SRC_EXT_REGEX: Pattern[str] = create_src_file_exts_str(
+            self.incl_suffixes
+        )
         # Intrinsic (re-loaded during initialize)
         (
             self.statements,
@@ -192,6 +198,7 @@ class LangServer:
         self.source_dirs.add(self.root_path)
 
         self._load_config_file()
+        update_recursion_limit(self.recursion_limit)
         self._resolve_globs_in_paths()
         self._config_logger(request)
         self._load_intrinsics()
@@ -204,7 +211,7 @@ class LangServer:
 
         # Initialize workspace
         self.workspace_init()
-        log.info(f"fortls - Fortran Language Server {__version__} Initialized")
+        log.info("fortls - Fortran Language Server %s Initialized", __version__)
         #
         server_capabilities = {
             "completionProvider": {
@@ -272,7 +279,7 @@ class LangServer:
 
     def serve_document_symbols(self, request: dict):
         def map_types(type, in_class: bool = False):
-            if type == 1:
+            if type in (1, 8):
                 return 2
             elif type in (2, 3):
                 if in_class:
@@ -300,7 +307,11 @@ class LangServer:
         # Add scopes to outline view
         test_output = []
         for scope in file_obj.ast.get_scopes():
-            if (scope.name[0] == "#") or (scope.get_type() == SELECT_TYPE_ID):
+            if (
+                not scope.name  # Skip empty strings
+                or scope.name.startswith("#")  # Skip comments
+                or scope.get_type() == SELECT_TYPE_ID  # Skip select types
+            ):
                 continue
             scope_tree = scope.FQSN.split("::")
             if len(scope_tree) > 2:
@@ -380,11 +391,14 @@ class LangServer:
         ):
             #
             def child_candidates(
-                scope, only_list: list = None, filter_public=True, req_abstract=False
-            ):
+                scope: Scope,
+                only_list: list = None,
+                filter_public=True,
+                req_abstract=False,
+            ) -> list[str]:
                 if only_list is None:
                     only_list = []
-                tmp_list = []
+                tmp_list: list[str] = []
                 # Filter children
                 nonly = len(only_list)
                 for child in scope.get_children(filter_public):
@@ -404,8 +418,8 @@ class LangServer:
                             tmp_list.append(child)
                 return tmp_list
 
-            var_list = []
-            use_dict = {}
+            var_list: list[str] = []
+            use_dict: dict[str, Use | Import] = {}
             for scope in scope_list:
                 var_list += child_candidates(
                     scope, filter_public=public_only, req_abstract=abstract_only
@@ -415,48 +429,58 @@ class LangServer:
                     use_dict = get_use_tree(scope, use_dict, self.obj_tree)
             # Look in found use modules
             rename_list = [None for _ in var_list]
+            import_var_list = []
             for use_mod, use_info in use_dict.items():
-                scope = self.obj_tree[use_mod][0]
-                only_list = use_info.only_list
-                if len(use_info.rename_map) > 0:
-                    only_list = [
-                        use_info.rename_map.get(only_name, only_name)
-                        for only_name in only_list
-                    ]
-                tmp_list = child_candidates(
-                    scope, only_list, req_abstract=abstract_only
-                )
-                # Setup renaming
-                if len(use_info.rename_map) > 0:
-                    rename_reversed = {
-                        value: key for (key, value) in use_info.rename_map.items()
-                    }
-                    for tmp_obj in tmp_list:
-                        var_list.append(tmp_obj)
-                        rename_list.append(
-                            rename_reversed.get(tmp_obj.name.lower(), None)
-                        )
-                else:
-                    var_list += tmp_list
-                    rename_list += [None for _ in tmp_list]
+                if type(use_info) is Use:
+                    scope = self.obj_tree[use_mod][0]
+                    only_list = use_info.rename()
+                    tmp_list = child_candidates(
+                        scope, only_list, req_abstract=abstract_only
+                    )
+                    # Setup renaming
+                    if use_info.rename_map:
+                        rename_reversed = {
+                            value: key for (key, value) in use_info.rename_map.items()
+                        }
+                        for tmp_obj in tmp_list:
+                            var_list.append(tmp_obj)
+                            rename_list.append(
+                                rename_reversed.get(tmp_obj.name.lower(), None)
+                            )
+                    else:
+                        var_list += tmp_list
+                        rename_list += [None for _ in tmp_list]
+                elif type(use_info) is Import:
+                    scope = use_info.scope
+                    # Add import candidates
+                    import_var_list += child_candidates(
+                        scope,
+                        only_list=use_info.only_list,
+                        req_abstract=abstract_only,
+                    )
+                    # We do not have renames so ignore
+
             # Add globals
             if inc_globals:
                 tmp_list = [obj[0] for (_, obj) in self.obj_tree.items()]
                 var_list += tmp_list + self.intrinsic_funs
                 rename_list += [None for _ in tmp_list + self.intrinsic_funs]
+            if import_var_list:
+                var_list = import_var_list
+                rename_list = [None for _ in import_var_list]
             # Filter by prefix if necessary
             if var_prefix == "":
                 return var_list, rename_list
             else:
-                tmp_list = []
-                tmp_rename = []
-                for (i, var) in enumerate(var_list):
-                    var_name = rename_list[i]
+                tmp_list: list[str] = []
+                tmp_rename: list[str] = []
+                for var, rename in zip(var_list, rename_list):
+                    var_name: str | None = rename
                     if var_name is None:
                         var_name = var.name
                     if var_name.lower().startswith(var_prefix):
                         tmp_list.append(var)
-                        tmp_rename.append(rename_list[i])
+                        tmp_rename.append(rename)
                 return tmp_list, tmp_rename
 
         def build_comp(
@@ -487,12 +511,21 @@ class LangServer:
             comp_obj["kind"] = map_types(candidate.get_type())
             if is_member and (comp_obj["kind"] == 3):
                 comp_obj["kind"] = 2
+            # Detail label shown above documentation, also shown when
+            # documentation is collapsed i.e. short form completions
             comp_obj["detail"] = candidate.get_desc()
             if call_sig is not None:
                 comp_obj["detail"] += " " + call_sig
-            doc_str, _ = candidate.get_hover()
-            if doc_str is not None:
-                comp_obj["documentation"] = doc_str
+            # Use the full markdown documentation
+            hover_msg: str = candidate.get_hover_md(long=True)
+            if hover_msg:
+                hover_msg: dict = {
+                    "kind": "markdown",
+                    "value": hover_msg.replace(
+                        "```{langid}", f"```{self.hover_language}", 1
+                    ),
+                }
+                comp_obj["documentation"] = hover_msg
             return comp_obj
 
         # Get parameters from request
@@ -641,7 +674,7 @@ class LangServer:
         candidate_list, rename_list = get_candidates(
             scope_list, var_prefix, include_globals, public_only, abstract_only, no_use
         )
-        for (i, candidate) in enumerate(candidate_list):
+        for candidate, rename in zip(candidate_list, rename_list):
             # Skip module names (only valid in USE)
             candidate_type = candidate.get_type()
             if type_mask[candidate_type]:
@@ -649,8 +682,8 @@ class LangServer:
             if req_callable and (not candidate.is_callable()):
                 continue
             #
-            name_replace = rename_list[i]
-            if candidate_type == INTERFACE_TYPE_ID:
+            name_replace = rename
+            if candidate_type == INTERFACE_TYPE_ID and not line_context == "mod_mems":
                 tmp_list = []
                 if name_replace is None:
                     name_replace = candidate.name
@@ -719,11 +752,18 @@ class LangServer:
             return None
         # Search in Preprocessor defined variables
         if def_name in def_file.pp_defs:
+            def_value = def_file.pp_defs.get(def_name)
+            def_arg_str = ""
+            if isinstance(def_value, tuple):
+                def_arg_str, def_value = def_value
+                def_arg_str = ", ".join([x.strip() for x in def_arg_str.split(",")])
+                def_arg_str = f"({def_arg_str})"
+
             var = Variable(
                 def_file.ast,
                 def_line + 1,
                 def_name,
-                f"#define {def_name} {def_file.pp_defs.get(def_name)}",
+                f"#define {def_name}{def_arg_str} {def_value}",
                 [],
             )
             return var
@@ -812,6 +852,17 @@ class LangServer:
                         return i
             return None
 
+        def replace_langid(params: list[dict]) -> list[dict]:
+            new_params = params[:]
+            for param in new_params:
+                if "documentation" not in param:
+                    continue
+                # Replace the first value of langid, when starting a code block
+                param["documentation"]["value"] = param["documentation"][
+                    "value"
+                ].replace("```{langid}", f"```{self.hover_language}", 1)
+            return params
+
         # Get parameters from request
         params: dict = request["params"]
         uri: str = params["textDocument"]["uri"]
@@ -879,6 +930,9 @@ class LangServer:
         label, doc_str, params = var_obj.get_signature()
         if label is None:
             return None
+        # Replace placeholder language id with Fortran ID
+        params = replace_langid(params)
+
         # Find current parameter by index or by
         # looking at last arg with optional name
         param_num = len(arg_strings) - 1
@@ -892,7 +946,8 @@ class LangServer:
             param_num = opt_num
         signature = {"label": label, "parameters": params}
         if doc_str is not None:
-            signature["documentation"] = doc_str
+            doc_str = doc_str.format(langid=self.hover_language)
+            signature["documentation"] = {"kind": "markdown", "value": doc_str}
         req_dict = {"signatures": [signature], "activeParameter": param_num}
         return req_dict
 
@@ -918,7 +973,7 @@ class LangServer:
         for filename, file_obj in file_set:
             file_refs = []
             # Search through file line by line
-            for (i, line) in enumerate(file_obj.contents_split):
+            for i, line in enumerate(file_obj.contents_split):
                 if len(line) == 0:
                     continue
                 # Skip comment lines
@@ -1008,7 +1063,7 @@ class LangServer:
                     return None
         all_refs, _ = self.get_all_references(def_obj, type_mem, file_obj=restrict_file)
         refs = []
-        for (filename, file_refs) in all_refs.items():
+        for filename, file_refs in all_refs.items():
             for ref in file_refs:
                 refs.append(
                     uri_json(path_to_uri(filename), ref[0], ref[1], ref[0], ref[2])
@@ -1035,38 +1090,17 @@ class LangServer:
         return None
 
     def serve_hover(self, request: dict):
-        def create_hover(string: str, highlight: bool):
-            if highlight:
-                return {"language": self.hover_language, "value": string}
-            else:
-                return string
-
-        def create_signature_hover():
-            sig_request = request.copy()
-            sig_result = self.serve_signature(sig_request)
-            try:
-                arg_id = sig_result.get("activeParameter")
-                if arg_id is not None:
-                    arg_info = sig_result["signatures"][0]["parameters"][arg_id]
-                    arg_doc = arg_info["documentation"]
-                    doc_split = arg_doc.find("\n !!")
-                    if doc_split < 0:
-                        arg_string = f"{arg_doc} :: {arg_info['label']}"
-                    else:
-                        arg_string = (
-                            f"{arg_doc[:doc_split]} :: "
-                            f"{arg_info['label']}{arg_doc[doc_split:]}"
-                        )
-                    return create_hover(arg_string, True)
-            except:
-                pass
+        def create_hover(string: str, docs: str | None):
+            # This does not account for Fixed Form Fortran, but it should be
+            # okay for 99% of cases
+            return fortran_md(string, docs).format(langid=self.hover_language)
 
         # Get parameters from request
         params: dict = request["params"]
         uri: str = params["textDocument"]["uri"]
         def_line: int = params["position"]["line"]
         def_char: int = params["position"]["character"]
-        path = path_from_uri(uri)
+        path: str = path_from_uri(uri)
         file_obj = self.workspace.get(path)
         if file_obj is None:
             return None
@@ -1075,41 +1109,46 @@ class LangServer:
         if var_obj is None:
             return None
         # Construct hover information
-        var_type = var_obj.get_type()
+        var_type: int = var_obj.get_type()
         hover_array = []
-        if var_type in (SUBROUTINE_TYPE_ID, FUNCTION_TYPE_ID):
-            hover_str, highlight = var_obj.get_hover(long=True)
-            hover_array.append(create_hover(hover_str, highlight))
+        if var_type in (
+            SUBROUTINE_TYPE_ID,
+            FUNCTION_TYPE_ID,
+            MODULE_TYPE_ID,
+            CLASS_TYPE_ID,
+        ):
+            hover_array.append(
+                var_obj.get_hover_md(long=True).replace(
+                    "```{langid}", f"```{self.hover_language}", 1
+                )
+            )
         elif var_type == INTERFACE_TYPE_ID:
             for member in var_obj.mems:
-                hover_str, highlight = member.get_hover(long=True)
+                hover_str, docs = member.get_hover(long=True)
                 if hover_str is not None:
-                    hover_array.append(create_hover(hover_str, highlight))
+                    hover_array.append(create_hover(hover_str, docs))
         elif var_type == VAR_TYPE_ID:
             # Unless we have a Fortran literal include the desc in the hover msg
             # See get_definition for an explanation about this default name
             if not var_obj.desc.startswith(FORTRAN_LITERAL):
-                hover_str, highlight = var_obj.get_hover()
-                hover_array.append(create_hover(hover_str, highlight))
+                hover_array.append(
+                    var_obj.get_hover_md(long=True).replace(
+                        "```{langid}", f"```{self.hover_language}", 1
+                    )
+                )
             # Hover for Literal variables
             elif var_obj.desc.endswith("REAL"):
-                hover_array.append(create_hover("REAL", True))
+                hover_array.append(create_hover("REAL", None))
             elif var_obj.desc.endswith("INTEGER"):
-                hover_array.append(create_hover("INTEGER", True))
+                hover_array.append(create_hover("INTEGER", None))
             elif var_obj.desc.endswith("LOGICAL"):
-                hover_array.append(create_hover("LOGICAL", True))
+                hover_array.append(create_hover("LOGICAL", None))
             elif var_obj.desc.endswith("STRING"):
                 hover_str = f"CHARACTER(LEN={len(var_obj.name)-2})"
-                hover_array.append(create_hover(hover_str, True))
+                hover_array.append(create_hover(hover_str, None))
 
-            # Include the signature if one is present e.g. if in an argument list
-            if self.hover_signature:
-                hover_str = create_signature_hover()
-                if hover_str is not None:
-                    hover_array.append(hover_str)
-        #
         if len(hover_array) > 0:
-            return {"contents": hover_array}
+            return {"contents": {"kind": "markdown", "value": "\n".join(hover_array)}}
         return None
 
     def serve_implementation(self, request: dict):
@@ -1134,6 +1173,10 @@ class LangServer:
             impl_obj = var_obj.link_obj
             if (impl_obj is not None) and (impl_obj.file_ast.file is not None):
                 return self._create_ref_link(impl_obj)
+        elif var_obj.parent.get_type() == INTERFACE_TYPE_ID:
+            # Find the first implementation of the interface
+            if var_obj.link_obj is not None:
+                return self._create_ref_link(var_obj.link_obj)
         return None
 
     def serve_rename(self, request: dict):
@@ -1349,7 +1392,7 @@ class LangServer:
                             return False, "File does not exist"  # Error during load
                 err_string, file_changed = file_obj.load_from_disk()
                 if err_string:
-                    log.error(f"{err_string} : {filepath}")
+                    log.error("%s : %s", err_string, filepath)
                     return False, err_string  # Error during file read
                 if not file_changed:
                     return False, None
@@ -1449,7 +1492,7 @@ class LangServer:
             result_obj = result.get()
             if isinstance(result_obj, str):
                 self.post_message(
-                    f"Initialization failed for file '{path}': {result_obj}"
+                    f"Initialization failed for file {path}: {result_obj}"
                 )
                 continue
             self.workspace[path] = result_obj
@@ -1508,7 +1551,7 @@ class LangServer:
             break
 
         try:
-            with open(config_path, "r") as jsonfile:
+            with open(config_path) as jsonfile:
                 config_dict = json5.load(jsonfile)
 
                 # Include and Exclude directories
@@ -1540,7 +1583,7 @@ class LangServer:
         self.source_dirs = set(config_dict.get("source_dirs", self.source_dirs))
         self.incl_suffixes = set(config_dict.get("incl_suffixes", self.incl_suffixes))
         # Update the source file REGEX
-        self.FORTRAN_SRC_EXT_REGEX = src_file_exts(self.incl_suffixes)
+        self.FORTRAN_SRC_EXT_REGEX = create_src_file_exts_str(self.incl_suffixes)
         self.excl_suffixes = set(config_dict.get("excl_suffixes", self.excl_suffixes))
 
     def _load_config_file_general(self, config_dict: dict) -> None:
@@ -1551,6 +1594,7 @@ class LangServer:
             "incremental_sync", self.incremental_sync
         )
         self.sync_type: int = 2 if self.incremental_sync else 1
+        self.recursion_limit = config_dict.get("recursion_limit", self.recursion_limit)
         self.sort_keywords = config_dict.get("sort_keywords", self.sort_keywords)
         self.disable_autoupdate = config_dict.get(
             "disable_autoupdate", self.disable_autoupdate
@@ -1784,6 +1828,22 @@ class LangServer:
         return False
 
 
+def update_recursion_limit(limit: int) -> None:
+    """Update the recursion limit of the Python interpreter
+
+    Parameters
+    ----------
+    limit : int
+        New recursion limit
+
+    Examples
+    --------
+    >>> update_recursion_limit(10000)
+    """
+    if limit != sys.getrecursionlimit():
+        sys.setrecursionlimit(limit)
+
+
 class JSONRPC2Error(Exception):
     def __init__(self, code, message, data=None):
         self.code = code
